#############################START LICENSE##########################################
# Copyright (C) 2019 Pedro Martinez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#############################END LICENSE##########################################


###########################################################################################
#
#   Script name: qc-lightrad
#
#   Description: This script performs automated EPID QC of the QC-3 phantom developed in Manitoba.
#   There are other tools out there that do this but generally the ROI are fixed whereas this script
#   aims to dynamically identify them using machine vision and the bibs in the phantom.
#
#   Example usage: python qc-jaws "/folder/"
#
#   Tool for calculating jaws junction shifts for linear accelerators. The script opens every DICOM
#   file in a given folder and creates a combined profile resulting from the superposition of the
#   two or more fields. It then detects the peak/through formed by the gap/overlap of the fields.
#   A window is then selected around this point and a Savitzky-Golay smoothing filter is then applied
#   to the combined profile. This new curve is then used iteratively to minimize the profile created
#   every time one of the profiles slide to close the gap or decrease the overlap. The profile will
#   achieve it greatest level of homogeneity when the dosimetric penumbra of both fields are
#   matched in space. The final result is the optimal calculation of the gap/overlap between the
#   two profiles. The software generates a pdf file with a summary of all the results.
#
#   The folder should contain:
#   2 X-jaws images
#   2 or 3 Y-jaws images
#   4 Field rotation images
#
#   Author: Pedro Martinez
#   pedro.enrique.83@gmail.com
#   5877000722
#   Date:2019-04-09
#
###########################################################################################

import argparse
import os

from pymedphys._imports import numpy as np
from pymedphys._imports import plt, pydicom
from tqdm import tqdm

from matplotlib.backends.backend_pdf import PdfPages
from scipy import signal

from .utils import minimize_field_rot as minFR
from .utils import minimize_junction_X as minX
from .utils import minimize_junction_Y as minY
from .utils import peak_find as pf
from .utils import peak_find_fieldrot as pffr
from .utils import utils as u


# axial visualization and scrolling
def multi_slice_viewer(volume, dx, dy):
    # remove_keymap_conflicts({'j', 'k'})
    fig, ax = plt.subplots()
    ax.volume = volume
    ax.index = volume.shape[2] // 2
    print(ax.index)
    extent = (0, 0 + (volume.shape[1] * dx), 0, 0 + (volume.shape[0] * dy))
    ax.imshow(volume[:, :, ax.index], extent=extent)
    ax.set_xlabel("x distance [mm]")
    ax.set_ylabel("y distance [mm]")
    ax.set_title("item=" + str(ax.index))
    fig.suptitle("Axial view", fontsize=16)
    fig.canvas.mpl_connect("key_press_event", process_key_axial)


def process_key_axial(event):
    fig = event.canvas.figure
    ax = fig.axes[0]
    if event.key == "j":
        previous_slice_axial(ax)
    elif event.key == "k":
        next_slice_axial(ax)
    ax.set_title("item=" + str(ax.index))
    fig.canvas.draw()


def previous_slice_axial(ax):
    volume = ax.volume
    ax.index = (ax.index - 1) % volume.shape[2]  # wrap around using %
    # print(ax.index, volume.shape[2])
    ax.images[0].set_array(volume[:, :, ax.index])


def next_slice_axial(ax):
    volume = ax.volume
    ax.index = (ax.index + 1) % volume.shape[2]
    # print(ax.index, volume.shape[2])
    ax.images[0].set_array(volume[:, :, ax.index])


# this subroutine will merge the two jaws into a single image and display a graph of the overlap
def merge_view_vert(volume, dx, dy):
    junctions = []

    # creating merged volume
    merge_vol = np.zeros((volume.shape[0], volume.shape[1]))

    # creating vector for processing along cols (one row)
    amplitude = np.zeros(
        (volume.shape[1], volume.shape[2])
    )  # 1 if it is vertical 0 if the bars are horizontal

    x = np.linspace(
        0, 0 + (volume.shape[1] * dx), volume.shape[1], endpoint=False
    )  # definition of the distance axis
    # x = np.arange(0,)#definition of the distance axis

    # merging the two images together
    ampl_resamp = np.zeros(((volume.shape[1]) * 10, volume.shape[2]))
    # amp_peak = np.zeros((volume.shape[1]) * 10)

    for item in tqdm(range(0, volume.shape[2])):
        merge_vol = merge_vol + volume[:, :, item]
        amplitude[:, item] = volume[int(volume.shape[0] / 2), :, item]
        ampl_resamp[:, item] = signal.resample(
            amplitude[:, item], int(len(amplitude)) * 10
        )  # resampling the amplitude vector
        # amp_peak = amp_peak + ampl_resamp[:, item] / volume.shape[2]

    fig, ax = plt.subplots(nrows=2, squeeze=True, figsize=(6, 8))

    extent = (0, 0 + (volume.shape[1] * dx), 0, 0 + (volume.shape[0] * dy))

    ax[0].imshow(merge_vol, extent=extent)
    # ax[0].set_aspect('equal', 'box')
    ax[0].set_xlabel("x distance [mm]")
    ax[0].set_ylabel("y distance [mm]")

    ax[1].plot(x, amplitude)
    ax[1].set_xlabel("x distance [mm]")
    ax[1].set_ylabel("amplitude")
    ax[1].legend()
    fig.suptitle("Merged volume", fontsize=16)

    # peaks, peak_type, peak_figs = peak_find(ampl_resamp, dx)
    peaks, peak_type, peak_figs = pf.peak_find(ampl_resamp, dx)
    junction_figs = minX.minimize_junction_X(ampl_resamp, peaks, peak_type, dx / 10)
    junctions.append(junction_figs)

    return fig, peak_figs, junctions


# this subroutine will merge the 4 jaws and analyze them
# each jaw is 6cm in length (60mm)
def merge_view_horz(volume, dx, dy):
    junctions = []

    # creating merged volume
    merge_vol = np.zeros((volume.shape[0], volume.shape[1]))

    # creating vector for processing along cols (one row)
    amplitude = np.zeros(
        (volume.shape[0], volume.shape[2])
    )  # 1 if it is vertical 0 if the bars are horizontal

    y = np.linspace(
        0, 0 + (volume.shape[0] * dy), volume.shape[0], endpoint=False
    )  # definition of the distance axis
    # x = np.arange(0,) #definition of the distance axis

    # merging the two images together
    ampl_resamp = np.zeros(((volume.shape[0]) * 10, volume.shape[2]))
    # amp_peak = np.zeros((volume.shape[0]) * 10)

    for item in tqdm(range(0, volume.shape[2])):
        merge_vol = merge_vol + volume[:, :, item]
        amplitude[:, item] = volume[:, int(volume.shape[1] / 2), item]
        ampl_resamp[:, item] = signal.resample(
            amplitude[:, item], int(len(amplitude)) * 10
        )  # resampling the amplitude vector
        # amp_peak = amp_peak + ampl_resamp[:, item] / volume.shape[2]

    fig, ax = plt.subplots(nrows=2, squeeze=True, figsize=(6, 8))

    extent = (0, 0 + (volume.shape[1] * dx), 0, 0 + (volume.shape[0] * dy))

    ax[0].imshow(merge_vol, extent=extent, aspect="auto")
    ax[0].set_xlabel("x distance [mm]")
    ax[0].set_ylabel("y distance [mm]")

    ax[1].plot(y, amplitude, label="Amplitude profile")
    ax[1].set_ylabel("amplitude")
    ax[1].set_xlabel("y distance [mm]")
    ax[1].legend()
    fig.suptitle("Merged volume", fontsize=16)

    # peaks, peak_type, peak_figs = peak_find(ampl_resamp, dy)
    peaks, peak_type, peak_figs = pf.peak_find(ampl_resamp, dy)
    # junction_figs = minimize_junction_Y(ampl_resamp, peaks, peak_type, dy / 10)
    junction_figs = minY.minimize_junction_Y(ampl_resamp, peaks, peak_type, dy / 10)
    junctions.append(junction_figs)

    return fig, peak_figs, junctions


# this subroutine will merge the 4 jaws and analyze the two upper and two lower pairs
# each jaw is 6cm in length (60mm)
def merge_view_filtrot(volume, dx, dy):

    volume_resort = np.copy(
        volume
    )  # this will hold the resorted volume 0 to 3 clockwise
    junctions_comb = []
    peaks_figs_comb = []

    # we need to create 4 matches

    # 0,1,2,3 will be tagged top left, top right, bottom right, bottom left
    for i in range(0, int(np.shape(volume)[2])):
        diag_stack = [
            0,
            0,
            0,
            0,
        ]  # we will sum along one direction whichever is biggest will tag the file
        for j in range(0, int(min([np.shape(volume)[0], np.shape(volume)[1]]) / 2)):
            # print('j=',j,int(np.shape(volume)[0] / 2)+j, int(np.shape(volume)[1] / 2)+j)
            diag_stack[0] = (
                diag_stack[0]
                + volume[
                    int(np.shape(volume)[0] / 2) - j,
                    int(np.shape(volume)[1] / 2) - j,
                    i,
                ]
            )
            diag_stack[1] = (
                diag_stack[1]
                + volume[
                    int(np.shape(volume)[0] / 2) - j,
                    int(np.shape(volume)[1] / 2) + j,
                    i,
                ]
            )
            diag_stack[2] = (
                diag_stack[2]
                + volume[
                    int(np.shape(volume)[0] / 2) + j,
                    int(np.shape(volume)[1] / 2) + j,
                    i,
                ]
            )
            diag_stack[3] = (
                diag_stack[3]
                + volume[
                    int(np.shape(volume)[0] / 2) + j,
                    int(np.shape(volume)[1] / 2) - j,
                    i,
                ]
            )

        volume_resort[:, :, np.argmax(diag_stack)] = volume[:, :, i]

    # creating merged volumes
    merge_vol = np.zeros((volume_resort.shape[0], volume_resort.shape[1]))

    # creating vector for processing (1 horizontal & 1 vertical)
    amplitude_horz = np.zeros(
        (volume_resort.shape[1], volume_resort.shape[2])
    )  # 1 if it is vertical 0 if the bars are horizontal
    amplitude_vert = np.zeros((volume_resort.shape[0], volume_resort.shape[2]))

    # y = np.linspace(0, 0 + (volume_resort.shape[0] * dy), volume_resort.shape[0],
    #                 endpoint=False)  # definition of the distance axis
    # x = np.linspace(0, 0 + (volume_resort.shape[1] * dy), volume_resort.shape[1],
    #                 endpoint=False)  # definition of the distance axis

    ampl_resamp_y1 = np.zeros(
        ((volume_resort.shape[0]) * 10, int(volume_resort.shape[2] / 2))
    )
    ampl_resamp_y2 = np.zeros(
        ((volume_resort.shape[0]) * 10, int(volume_resort.shape[2] / 2))
    )

    ampl_resamp_x1 = np.zeros(
        ((volume_resort.shape[1]) * 10, int(volume_resort.shape[2] / 2))
    )
    ampl_resamp_x2 = np.zeros(
        ((volume_resort.shape[1]) * 10, int(volume_resort.shape[2] / 2))
    )

    amplitude_horz[:, 0] = volume_resort[
        int(volume_resort.shape[0] / 3.25), :, 0
    ]  # for profile 1
    amplitude_horz[:, 1] = volume_resort[
        int(volume_resort.shape[0] / 3.25), :, 1
    ]  # for profile 1
    amplitude_horz[:, 3] = volume_resort[
        int(volume_resort.shape[0]) - int(volume_resort.shape[0] / 3.25), :, 2
    ]  # the numbers here are reversed because we are going to slide the second graph (the overlay) to minimize the error  #for profile 2
    amplitude_horz[:, 2] = volume_resort[
        int(volume_resort.shape[0]) - int(volume_resort.shape[0] / 3.25), :, 3
    ]

    amplitude_vert[:, 0] = volume_resort[
        :, int(volume_resort.shape[1]) - int(volume_resort.shape[1] / 2.8), 1
    ]  # the numbers here are reversed because we are going to slide the second graph (the overlay) to minimize the error #for profile 3
    amplitude_vert[:, 1] = volume_resort[
        :, int(volume_resort.shape[1]) - int(volume_resort.shape[1] / 2.8), 2
    ]
    amplitude_vert[:, 3] = volume_resort[
        :, int(volume_resort.shape[1] / 2.8), 3
    ]  # for profile 4
    amplitude_vert[:, 2] = volume_resort[:, int(volume_resort.shape[1] / 2.8), 0]

    plt.figure()
    for item in tqdm(range(0, int(volume.shape[2] / 2))):
        merge_vol = merge_vol + volume[:, :, item]

        data_samp = amplitude_vert[:, item]
        ampl_resamp_y1[:, item] = signal.resample(
            data_samp, int(np.shape(amplitude_vert)[0]) * 10
        )
        data_samp = amplitude_horz[:, item]
        ampl_resamp_x1[:, item] = signal.resample(
            data_samp, int(np.shape(amplitude_horz)[0]) * 10
        )

    for item in tqdm(range(int(volume.shape[2] / 2), volume.shape[2])):
        merge_vol = merge_vol + volume[:, :, item]
        data_samp = amplitude_vert[:, item]
        ampl_resamp_y2[:, item - int(volume.shape[2] / 2)] = signal.resample(
            data_samp, int(np.shape(amplitude_vert)[0]) * 10
        )
        data_samp = amplitude_horz[:, item]
        ampl_resamp_x2[:, item - int(volume.shape[2] / 2)] = signal.resample(
            data_samp, int(np.shape(amplitude_horz)[0]) * 10
        )

    fig, ax = plt.subplots(ncols=1, nrows=1, squeeze=True, figsize=(6, 8))

    extent = (0, 0 + (volume.shape[1] * dx), 0, 0 + (volume.shape[0] * dy))

    ax.imshow(merge_vol, extent=extent, aspect="auto")
    ax.set_aspect("equal", "box")
    ax.set_xlabel("x distance [mm]")
    ax.set_ylabel("y distance [mm]")
    fig.suptitle("Merged volume", fontsize=16)

    ax.hlines(dy * int(volume_resort.shape[0] / 3.25), 0, dx * volume_resort.shape[1])
    ax.text(
        dx * int(volume_resort.shape[1] / 2.25),
        dy * int(volume_resort.shape[0] / 3),
        "Profile 2",
    )

    ax.hlines(
        dy * int(volume_resort.shape[0]) - dy * int(volume_resort.shape[0] / 3.25),
        0,
        dx * volume_resort.shape[1],
    )
    ax.text(
        dx * int(volume_resort.shape[1] / 2.25),
        dy * int(volume_resort.shape[0]) - dy * int(volume_resort.shape[0] / 3.5),
        "Profile 1",
    )

    ax.vlines(dx * int(volume_resort.shape[1] / 2.8), 0, dy * volume_resort.shape[0])
    ax.text(
        dx * int(volume_resort.shape[1] / 3.1),
        dy * int(volume_resort.shape[0] / 1.8),
        "Profile 4",
        rotation=90,
    )

    ax.vlines(
        dx * int(volume_resort.shape[1]) - dx * int(volume_resort.shape[1] / 2.8),
        0,
        dy * volume_resort.shape[0],
    )
    ax.text(
        dx * int(volume_resort.shape[1]) - dx * int(volume_resort.shape[1] / 2.9),
        dy * int(volume_resort.shape[0] / 1.8),
        "Profile 3",
        rotation=90,
    )
    # plt.show()

    peaks, peak_type, peak_figs = pffr.peak_find_fieldrot(
        ampl_resamp_x1, dx, "Profile 1"
    )
    junction_figs = minFR.minimize_junction_fieldrot(
        ampl_resamp_x1, peaks, peak_type, dx / 10, "Profile 1"
    )
    peaks_figs_comb.append(peak_figs)
    junctions_comb.append(junction_figs)

    peaks, peak_type, peak_figs = pffr.peak_find_fieldrot(
        ampl_resamp_x2, dx, "Profile 2"
    )
    junction_figs = minFR.minimize_junction_fieldrot(
        ampl_resamp_x2, peaks, peak_type, dx / 10, "Profile 2"
    )
    peaks_figs_comb.append(peak_figs)
    junctions_comb.append(junction_figs)

    peaks, peak_type, peak_figs = pffr.peak_find_fieldrot(
        ampl_resamp_y1, dy, "Profile 3"
    )
    junction_figs = minFR.minimize_junction_fieldrot(
        ampl_resamp_y1, peaks, peak_type, dy / 10, "Profile 3"
    )
    peaks_figs_comb.append(peak_figs)
    junctions_comb.append(junction_figs)

    peaks, peak_type, peak_figs = pffr.peak_find_fieldrot(
        ampl_resamp_y2, dy, "Profile 4"
    )
    junction_figs = minFR.minimize_junction_fieldrot(
        ampl_resamp_y2, peaks, peak_type, dy / 10, "Profile 4"
    )
    peaks_figs_comb.append(peak_figs)
    junctions_comb.append(junction_figs)

    return fig, peaks_figs_comb, junctions_comb


# this routine anlyzes the volume and autodetect the images and categorizes them for different tests
def image_analyze(volume, i_opt):
    xfield = []
    yfield = []
    rotfield = []

    if i_opt.startswith(("y", "yeah", "yes")):
        kx = 0
        ky = 0
        krot = 0
        for item in range(0, volume.shape[2]):
            stack1 = np.sum(
                volume[
                    int(
                        np.shape(volume)[0] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[0] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    int(
                        np.shape(volume)[1] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[1] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    item,
                ],
                axis=0,
            )
            maxstack1 = np.amax(stack1)

            # stack2 = np.sum(volume[:, :, item], axis=1)
            stack2 = np.sum(
                volume[
                    int(
                        np.shape(volume)[0] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[0] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    int(
                        np.shape(volume)[1] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[1] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    item,
                ],
                axis=1,
            )
            maxstack2 = np.amax(stack2)

            if maxstack2 / maxstack1 > 1.1:  # It is a Y field folder
                if ky == 0:
                    yfield = volume[:, :, item]
                    yfield = yfield[:, :, np.newaxis]
                else:
                    volappend = volume[:, :, item]
                    yfield = np.append(yfield, volappend[:, :, np.newaxis], axis=2)
                ky = ky + 1
            elif maxstack2 / maxstack1 < 0.9:  # It is a X field folder
                if kx == 0:
                    xfield = volume[:, :, item]
                    xfield = xfield[:, :, np.newaxis]
                else:
                    # xfield=xfield[:,:,np.newaxis]
                    volappend = volume[:, :, item]
                    xfield = np.append(xfield, volappend[:, :, np.newaxis], axis=2)
                kx = kx + 1
            else:  # It is a field rotation folder
                if krot == 0:
                    rotfield = volume[:, :, item]
                    rotfield = rotfield[:, :, np.newaxis]
                else:
                    # rotfield = rotfield[:, :, np.newaxis]
                    volappend = volume[:, :, item]
                    rotfield = np.append(rotfield, volappend[:, :, np.newaxis], axis=2)
                krot = krot + 1

    else:
        kx = 0
        ky = 0
        krot = 0
        for item in range(0, volume.shape[2]):
            stack1 = np.sum(
                volume[
                    int(
                        np.shape(volume)[0] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[0] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    int(
                        np.shape(volume)[1] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[1] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    item,
                ],
                axis=0,
            )
            maxstack1 = np.amax(stack1)

            # stack2 = np.sum(volume[:, :, item], axis=1)
            stack2 = np.sum(
                volume[
                    int(
                        np.shape(volume)[0] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[0] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    int(
                        np.shape(volume)[1] / 2
                        - np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ) : int(
                        np.shape(volume)[1] / 2
                        + np.amin([np.shape(volume)[0], np.shape(volume)[1]]) / 2
                    ),
                    item,
                ],
                axis=1,
            )
            maxstack2 = np.amax(stack2)

            if maxstack2 / maxstack1 > 1.5:  # It is a Y field folder
                if ky == 0:
                    yfield = volume[:, :, item]
                    yfield = yfield[:, :, np.newaxis]
                else:
                    volappend = volume[:, :, item]
                    yfield = np.append(yfield, volappend[:, :, np.newaxis], axis=2)
                ky = ky + 1
            elif maxstack2 / maxstack1 < 0.5:  # It is a X field folder
                if kx == 0:
                    xfield = volume[:, :, item]
                    xfield = xfield[:, :, np.newaxis]
                else:
                    # xfield=xfield[:,:,np.newaxis]
                    volappend = volume[:, :, item]
                    xfield = np.append(xfield, volappend[:, :, np.newaxis], axis=2)
                kx = kx + 1
            else:  # It is a field rotation folder
                if krot == 0:
                    rotfield = volume[:, :, item]
                    rotfield = rotfield[:, :, np.newaxis]
                else:
                    # rotfield = rotfield[:, :, np.newaxis]
                    volappend = volume[:, :, item]
                    rotfield = np.append(rotfield, volappend[:, :, np.newaxis], axis=2)
                krot = krot + 1

    return xfield, yfield, rotfield


# this routine anlyzes the volume and autodetect what type of analysis to carry on (x, Y, Field Rot)
def folder_analyze(volume):
    for item in range(0, volume.shape[2]):
        stack1 = np.sum(volume[:, :, item], axis=0)
        maxstack1 = np.max(stack1)

        stack2 = np.sum(volume[:, :, item], axis=1)
        maxstack2 = np.max(stack2)

        if maxstack2 / maxstack1 > 1.5:  # It is a Y field folder
            field = 2
        elif maxstack2 / maxstack1 < 0.5:  # It is a X field folder
            field = 1
        else:
            field = 3  # It is a field rotation folder

        return field


def read_dicom3D(direc, i_option):
    # item = 0
    for subdir, dirs, files in os.walk(direc):  # pylint: disable = unused-variable
        k = 0
        for file in tqdm(sorted(files)):
            # print('filename=', file)
            if os.path.splitext(file)[1] == ".dcm":
                dataset = pydicom.dcmread(direc + file)
                if k == 0:
                    ArrayDicom = np.zeros(
                        (dataset.Rows, dataset.Columns, 0),
                        dtype=dataset.pixel_array.dtype,
                    )
                    tmp_array = dataset.pixel_array
                    if i_option.startswith(("y", "yeah", "yes")):
                        max_val = np.amax(tmp_array)
                        tmp_array = tmp_array / max_val
                        min_val = np.amin(tmp_array)
                        tmp_array = tmp_array - min_val
                        tmp_array = 1 - tmp_array  # inverting the range

                        # min_val = np.amin(tmp_array)  # normalizing
                        # tmp_array = tmp_array - min_val
                        # tmp_array = tmp_array / (np.amax(tmp_array))
                        tmp_array = u.norm01(tmp_array)
                    else:
                        # min_val = np.amin(tmp_array)
                        # tmp_array = tmp_array - min_val
                        # tmp_array = tmp_array / (np.amax(tmp_array))
                        tmp_array = u.norm01(tmp_array)  # just normalize
                    ArrayDicom = np.dstack((ArrayDicom, tmp_array))
                    # print("item thickness [mm]=", dataset.SliceThickness)
                    SID = dataset.RTImageSID
                    dx = 1 / (SID * (1 / dataset.ImagePlanePixelSpacing[0]) / 1000)
                    dy = 1 / (SID * (1 / dataset.ImagePlanePixelSpacing[1]) / 1000)
                    print("pixel spacing row [mm]=", dx)
                    print("pixel spacing col [mm]=", dy)
                else:
                    tmp_array = dataset.pixel_array
                    if i_option.startswith(("y", "yeah", "yes")):
                        max_val = np.amax(tmp_array)
                        tmp_array = tmp_array / max_val
                        min_val = np.amin(tmp_array)
                        tmp_array = tmp_array - min_val
                        tmp_array = 1 - tmp_array  # inverting the range

                        # min_val = np.amin(tmp_array)  # normalizing
                        # tmp_array = tmp_array - min_val
                        # tmp_array = tmp_array / (np.amax(tmp_array))
                        tmp_array = u.norm01(tmp_array)
                    else:
                        # min_val = np.amin(tmp_array)
                        # tmp_array = tmp_array - min_val
                        # tmp_array = tmp_array / (np.amax(tmp_array))  # just normalize
                        tmp_array = u.norm01(tmp_array)
                    ArrayDicom = np.dstack((ArrayDicom, tmp_array))
            k = k + 1

    xfield, yfield, rotfield = image_analyze(ArrayDicom, i_option)

    multi_slice_viewer(ArrayDicom, dx, dy)

    if np.shape(xfield)[2] == 2:
        fig, peak_figs, junctions_figs = merge_view_vert(xfield, dx, dy)
        with PdfPages(direc + "jaws_X_report.pdf") as pdf:
            pdf.savefig(fig)
            # for i in range(0, len(peak_figs)):
            for _, f in enumerate(peak_figs):
                pdf.savefig(f)

            # for i in range(0, len(junctions_figs)):
            for _, f in enumerate(junctions_figs):
                pdf.savefig(f)

            plt.close()

    else:
        print(
            "X jaws data analysis not completed please verify that you have two X jaws images. For more information see manual."
        )

    if np.shape(yfield)[2] == 4:
        fig, peak_figs, junctions_figs = merge_view_horz(yfield, dx, dy)
        # print('peak_figs********************************************************=', len(peak_figs),peak_figs)
        with PdfPages(direc + "jaws_Y_report.pdf") as pdf:
            pdf.savefig(fig)
            # for i in range(0, len(peak_figs)):
            for _, f in enumerate(peak_figs):
                pdf.savefig(f)

            for _, f in enumerate(junctions_figs):
                pdf.savefig(f)

            plt.close()

    else:
        print(
            "Y jaws data analysis not completed please verify that you have four Y jaws images. For more information see manual."
        )

    if np.shape(rotfield)[2] == 4:
        fig, peak_figs, junctions_figs = merge_view_filtrot(rotfield, dx, dy)

        with PdfPages(direc + "jaws_FR_report.pdf") as pdf:
            pdf.savefig(fig)
            for _, f in enumerate(peak_figs):
                pdf.savefig(f)

            for _, f in enumerate(junctions_figs):
                pdf.savefig(f)

            plt.close()

    else:
        print(
            "Field rotation data analysis not completed please verify that you have four field rotation images. For more information see manual."
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--directory", help="path to folder")
    args = parser.parse_args()

    while True:  # example of infinite loops using try and except to catch only numbers
        line = input("Are these files from a clinac [yes(y)/no(n)]> ")
        try:
            ##        if line == 'done':
            ##            break
            ioption = str(line.lower())
            if ioption.startswith(("y", "yeah", "yes", "n", "no", "nope")):
                break

        except:  # pylint: disable = bare-except
            print("Please enter a valid option:")

    if args.directory:
        dirname = args.directory
        read_dicom3D(dirname, ioption)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np
from pymedphys._imports import scipy


def gaussian_cdf(x, mu=0, sig=1):
    x = np.array(x, copy=False)
    return 0.5 * (1 + scipy.special.erf((x - mu) / (sig * np.sqrt(2))))


def scaled_penumbra_sig(profile_shoulder_edge=0.8):
    sig = 1 / (2 * np.sqrt(2) * scipy.special.erfinv(profile_shoulder_edge * 2 - 1))

    return sig


def create_profile_function(centre, field_width, penumbra_width):
    sig = scaled_penumbra_sig() * penumbra_width
    mu = [centre - field_width / 2, centre + field_width / 2]

    def profile(x):
        x = np.array(x, copy=False)
        return gaussian_cdf(x, mu[0], sig) * gaussian_cdf(-x, -mu[1], sig)

    return profile


def rotate_coords(x, y, theta):
    x_prime = x * np.cos(theta) + y * np.sin(theta)
    y_prime = -x * np.sin(theta) + y * np.cos(theta)

    return x_prime, y_prime


def create_rectangular_field_function(centre, side_lengths, penumbra_width, rotation=0):
    width_profile = create_profile_function(0, side_lengths[0], penumbra_width)
    length_profile = create_profile_function(0, side_lengths[1], penumbra_width)

    theta = -rotation / 180 * np.pi

    def field(x, y):
        x = np.array(x, copy=False)
        y = np.array(y, copy=False)
        x_shifted = x - centre[0]
        y_shifted = y - centre[1]
        x_rotated, y_rotated = rotate_coords(x_shifted, y_shifted, theta)

        return width_profile(x_rotated) * length_profile(y_rotated)

    return field


def create_square_field_function(centre, side_length, penumbra_width, rotation=0):

    side_lengths = [side_length, side_length]
    return create_rectangular_field_function(
        centre, side_lengths, penumbra_width, rotation
    )



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime

from pymedphys._imports import dateutil

from pymedphys._mosaiq.delivery import get_mosaiq_delivery_details
from pymedphys._trf.decode.header import decode_header_from_file

# TODO Make the field identification run one whole day at a time, searching
# for all logfiles on that day and then attempting to align the logfiles
# to the full day's machine schedule.
# This will allow for the code to be more flexible on time differences and
# result in greater robustness to time differences between TCS and Mosaiq.


def date_convert(date, timezone):
    """Converts logfile UTC date to the provided timezone.
    The date is formatted to match the syntax required by Microsoft SQL."""

    from_timezone = dateutil.tz.gettz("UTC")
    to_timezone = dateutil.tz.gettz(timezone)

    utc_datetime = datetime.strptime(date, "%y/%m/%d %H:%M:%S Z").replace(
        tzinfo=from_timezone
    )
    local_time = utc_datetime.astimezone(to_timezone)
    mosaiq_string_time = local_time.strftime("%Y-%m-%d %H:%M:%S")
    path_string_time = local_time.strftime("%Y-%m-%d_%H%M%S")

    return mosaiq_string_time, path_string_time


def identify_logfile(cursor, filepath, timezone):
    header = decode_header_from_file(filepath)

    if header.field_label == "":
        raise ValueError("No field label in logfile")

    mosaiq_string_time, _ = date_convert(header.date, timezone)

    delivery_details = get_mosaiq_delivery_details(
        cursor,
        header.machine,
        mosaiq_string_time,
        header.field_label,
        header.field_name,
    )

    return delivery_details



# Copyright (C) 2020 Simon Biggs
# Copyright 2001-2017 by Vinay Sajip. All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of Vinay Sajip
# not be used in advertising or publicity pertaining to distribution
# of the software without specific, written prior permission.
# VINAY SAJIP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# VINAY SAJIP BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
# ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Code vendored and modified from:
# https://github.com/python/cpython/blob/a667e1c66a/Lib/logging/__init__.py#L1957-L1961

import logging
import sys

from pymedphys import _config

_patch_applied = False
_basicConfig = logging.basicConfig


def apply_logging_patch():
    if sys.version_info.major > 3 or (
        sys.version_info.major == 3 and sys.version_info.minor >= 8
    ):
        return

    global _patch_applied  # pylint: disable = global-statement

    if not _config.is_cli:
        raise ValueError(
            "This patch globally adjusts the logging module. This patch is "
            "not to be used within pymedphys while it is in library mode "
            "as monkey patches on libraries/modules external to pymedphys "
            "can be exceptionally confusing to a library user. "
            "This is only to be used when pymedphys is being called via "
            "the pymedphys CLI."
        )

    if _patch_applied:
        raise ValueError("This patch has already been applied.")

    def basicConfig(**kwargs):
        force = kwargs.pop("force", False)
        if force:
            for h in logging.root.handlers[:]:  # pylint: disable = no-member
                logging.root.removeHandler(h)  # pylint: disable = no-member
                h.close()

        _basicConfig(**kwargs)

    logging.basicConfig = basicConfig
    _patch_applied = True



# Copyright (C) 2018 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Test end to end conversion agrees."""

import pathlib
import tempfile

import pytest

import numpy as np
import pandas as pd

import pymedphys
from pymedphys._trf.decode.trf2csv import trf2csv

# TODO need to include header test


def compare_reference_to_converted(reference_dataframe, converted_dataframe):
    np.testing.assert_array_equal(
        reference_dataframe.columns,
        converted_dataframe.columns,
        "Columns should be equal.",
    )

    np.testing.assert_array_equal(
        reference_dataframe.index, converted_dataframe.index, "Index should be equal."
    )

    if not np.all(reference_dataframe.values == converted_dataframe.values):
        for column in reference_dataframe.columns:
            np.testing.assert_array_equal(
                reference_dataframe[column].values,
                converted_dataframe[column].values,
                "The {} column should be equal".format(column),
            )


def convert_and_check_against_baseline(filepath: pathlib.Path, output_directory):
    baseline_csv_file = filepath.parent.joinpath(filepath.stem + "_table.csv")
    baseline_dataframe = pd.read_csv(baseline_csv_file, index_col=0)

    convert_and_check(filepath, output_directory, baseline_dataframe)


def convert_and_check_against_reference(filepath: pathlib.Path, output_directory):
    reference_csv_file = filepath.with_suffix(".csv")
    reference_dataframe = pd.read_csv(reference_csv_file, skiprows=9, index_col=0)
    del reference_dataframe[reference_dataframe.columns[-1]]

    convert_and_check(filepath, output_directory, reference_dataframe)


def convert_and_check(filepath: pathlib.Path, output_directory, reference_dataframe):
    _, table_filepath = trf2csv(filepath, output_directory=output_directory)
    converted_dataframe = pd.read_csv(table_filepath, index_col=0)

    compare_reference_to_converted(reference_dataframe, converted_dataframe)


@pytest.mark.slow
def test_conversions():
    data_paths = pymedphys.zip_data_paths("trf-references-and-baselines.zip")

    files_with_references = [
        path
        for path in data_paths
        if path.parent.name == "with_reference" and path.suffix == ".trf"
    ]

    assert len(files_with_references) >= 5

    files_without_references = [
        path
        for path in data_paths
        if path.parent.name == "with_baseline" and path.suffix == ".trf"
    ]

    assert len(files_without_references) >= 4

    with tempfile.TemporaryDirectory() as output_directory:
        for filepath in files_with_references:
            convert_and_check_against_reference(filepath, output_directory)

        for filepath in files_without_references:
            convert_and_check_against_baseline(filepath, output_directory)



# Copyright (C) 2018 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable=C0103,C1801


"""Regression testing of the single control point function.
"""

import numpy as np

from pymedphys._metersetmap.metersetmap import calc_single_control_point

MLC = np.array([[[1, 1], [2, 2]], [[2, 2], [3, 3]]])

JAW = np.array([[1.5, 1.2], [1.5, 1.2]])

LEAF_PAIR_WIDTHS = [2, 2]

REFERENCE_METERSETMAP = [
    [0.0, 0.07, 0.43, 0.5, 0.43, 0.07, 0.0],
    [0.0, 0.14, 0.86, 1.0, 0.86, 0.14, 0.0],
    [0.14, 0.86, 1.0, 1.0, 1.0, 0.86, 0.14],
    [0.03, 0.17, 0.2, 0.2, 0.2, 0.17, 0.03],
]


def test_partial_jaws():
    """Partial jaw location should give a fractional result.
    """
    _, metersetmap = calc_single_control_point(
        MLC, JAW, leaf_pair_widths=LEAF_PAIR_WIDTHS
    )

    assert np.allclose(np.round(metersetmap, 2), REFERENCE_METERSETMAP)



"""Module docstring."""

from . import dicom, electronfactors, mosaiq, metersetmap, mudensity, trf, wlutz
from ._data import data_path, zenodo_data_paths, zip_data_paths
from ._delivery import Delivery
from ._gamma.implementation.shell import gamma_shell as gamma
from ._trf.decode import read_trf as _read_trf
from ._vendor.deprecated import deprecated as _deprecated
from ._version import __version__, version_info

_read_trf.__name__ = "pymedphys.read_trf"
read_trf = _deprecated(reason="This has been replaced with `pymedphys.trf.read`")(
    _read_trf
)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import lzma
import operator
import os
from pathlib import Path

import pytest

import numpy as np
import pandas as pd

import pydicom

from pymedphys._data import download
from pymedphys.dicom import depth_dose, profile
from pymedphys.experimental.fileformats import load_mephysto_directory

HERE = Path(__file__).parent.resolve()
DATA_DIR = HERE.joinpath("data")

DICOM_DIR = DATA_DIR.joinpath("DICOM")
DICOM_DOSE_FILEPATHS = {"05x05": "06MV_05x05.dcm.xz", "10x10": "06MV_10x10.dcm.xz"}
DICOM_PLAN_FILEPATH = "06MV_plan.dcm"

MEASUREMENTS_DIR = DATA_DIR.joinpath("measurements")
BASELINES_DIR = DATA_DIR.joinpath("baselines")

CREATE_BASELINE = False
BASELINE_FILEPATH = os.path.join(BASELINES_DIR, "dicom_dose_profiles.json")

# pylint: disable = redefined-outer-name


@pytest.fixture
def loaded_doses():
    doses = {}

    for key, filepath in DICOM_DOSE_FILEPATHS.items():

        resolved_filepath = str(
            download.get_file_within_data_zip("tps_compare_dicom_files.zip", filepath)
        )

        with lzma.open(resolved_filepath) as a_file:
            doses[key] = pydicom.dcmread(a_file, force=True)

    return doses


@pytest.fixture
def loaded_plan():
    plan = pydicom.read_file(
        str(
            download.get_file_within_data_zip(
                "tps_compare_dicom_files.zip", DICOM_PLAN_FILEPATH
            )
        ),
        force=True,
    )

    return plan


@pytest.mark.pydicom
def test_bulk_compare(loaded_doses, loaded_plan):
    absolute_dose_table = pd.read_csv(
        MEASUREMENTS_DIR.joinpath("AbsoluteDose.csv"), index_col=0
    )
    absolute_dose = absolute_dose_table["d10 @ 90 SSD"]["6 MV"]

    output_factors = pd.read_csv(
        MEASUREMENTS_DIR.joinpath("OutputFactors.csv"), index_col=0
    )

    absolute_doses = {
        key: output_factors[key]["6 MV"] * absolute_dose
        for key in output_factors.columns
    }

    absolute_scans_per_field = load_mephysto_directory(
        MEASUREMENTS_DIR, r"06MV_(\d\dx\d\d)\.mcc", absolute_doses, 100
    )

    getter = operator.itemgetter("displacement", "dose")

    for key, absolute_scans in absolute_scans_per_field.items():
        dose_dataset = loaded_doses[key]

        depths, meas_dose = getter(absolute_scans["depth_dose"])
        tps_dose = depth_dose(depths, dose_dataset, loaded_plan) / 10
        diff = tps_dose - meas_dose

        assert np.abs(np.mean(diff)) <= 0.02
        assert np.std(diff) <= 0.05


@pytest.mark.pydicom
def test_baseline_profiles(loaded_doses, loaded_plan):
    baselines = {}

    displacements = list(range(-100, 110, 10))
    depths = list(range(0, 310, 10))

    for key, dose_dataset in loaded_doses.items():
        baselines[key] = {}

        extracted_dose = depth_dose(depths, dose_dataset, loaded_plan)
        rounded_result = np.around(extracted_dose, decimals=2)
        baselines[key]["depth"] = rounded_result.tolist()

        for direction in ["inplane", "crossplane"]:
            baselines[key][direction] = {}
            for depth in [50, 100]:
                extracted_dose = profile(
                    displacements, depth, direction, dose_dataset, loaded_plan
                )
                rounded_result = np.around(extracted_dose, decimals=2)
                baselines[key][direction][str(depth)] = rounded_result.tolist()

    if CREATE_BASELINE:
        with open(BASELINE_FILEPATH, "w") as a_file:
            json.dump(baselines, a_file)

    else:
        with open(BASELINE_FILEPATH, "r") as a_file:
            baseline_result = json.load(a_file)

        assert baseline_result == baselines



from pymedphys._base.delivery import DeliveryBase
from pymedphys._dicom.delivery import DeliveryDicom
from pymedphys._icom.delivery import DeliveryIcom
from pymedphys._monaco.delivery import DeliveryMonaco
from pymedphys._mosaiq.delivery import DeliveryMosaiq
from pymedphys._metersetmap.delivery import DeliveryMetersetMap
from pymedphys._trf.decode.delivery import DeliveryLogfile


class Delivery(  # type: ignore
    DeliveryMosaiq,
    DeliveryMetersetMap,
    DeliveryLogfile,
    DeliveryDicom,
    DeliveryMonaco,
    DeliveryIcom,
    DeliveryBase,
):
    pass



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from pathlib import Path

from pymedphys._imports import numpy as np
from pymedphys._imports import plt

import imageio
from scipy.optimize import basinhopping

from .optical_density import calc_net_od

DEFAULT_CAL_STRING_END = " cGy.tif"


def create_dose_function(net_od, dose):
    net_od = np.array(net_od, copy=False)
    dose = np.array(dose, copy=False)

    to_minimise = create_to_minimise(net_od, dose)
    result = basinhopping(to_minimise, [np.max(dose) / np.max(net_od), 1, 1])

    return create_cal_fit(*result.x)


def create_cal_fit(a, b, n):
    def cal_fit(net_od):
        net_od = np.array(net_od, copy=False)
        return a * net_od + b * net_od ** n

    return cal_fit


def create_to_minimise(net_od, dose):
    def to_minimise(x):
        a, b, n = x

        cal_fit = create_cal_fit(a, b, n)
        return np.sum((cal_fit(net_od) - dose) ** 2)

    return to_minimise


def calc_calibration_points(
    prescans, postscans, alignments=None, figures=False, pixel_trim=0
):
    """Returns calibration points based on dictionaries of prescans and postscans.

    The key of the dictionaries of images is to represent the dose calibration
    point. If the key cannot be converted into a float that image will be
    ignored.
    """
    keys = prescans.keys()
    assert keys == postscans.keys()

    calibration_points = {}

    if alignments is None:
        alignments = {key: None for key in keys}

    for key in keys:
        try:
            dose_value = float(key)
        except ValueError:
            warnings.warn(
                "{} does not appear to be a calibration image key. This will "
                "be skipped."
            )
            continue

        net_od, alignment = calc_net_od(
            prescans[key], postscans[key], alignment=alignments[key]
        )

        if pixel_trim != 0:
            trim_ref = (slice(pixel_trim, -pixel_trim), slice(pixel_trim, -pixel_trim))
            net_od = net_od[trim_ref]

        if figures:
            plt.figure()
            plt.imshow(net_od)
            plt.show()

        calibration_points[dose_value] = np.median(net_od)
        alignments[key] = alignment

    return calibration_points, alignments


def load_cal_scans(path, cal_string_end=DEFAULT_CAL_STRING_END):
    path = Path(path)

    cal_pattern = "*" + cal_string_end
    filepaths = path.glob(cal_pattern)

    calibrations = {
        float(path.name.rstrip(cal_string_end)): load_image(path) for path in filepaths
    }

    return calibrations


def load_image(path):
    return imageio.imread(path)



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import functools
import json
import pathlib

from pymedphys._imports import numpy as np
from pymedphys._imports import pydicom
from pymedphys._imports import tensorflow as tf

from pymedphys._data import download

from . import filtering, indexing, mask


# @functools.lru_cache()
def get_dataset_metadata():
    release_url = "https://github.com/pymedphys/data/releases/download/structure-dicom"
    dicom_zip_url_pattern = f"{release_url}/" + "{dicom_type}.{uid}_Anonymised.zip"
    mappings_url = f"{release_url}/mappings.zip"

    data_download_root = pathlib.Path("auto-segmentation-dicom")

    save_filename = data_download_root.joinpath(get_filename_from_url(mappings_url))

    mappings_paths = download.zip_data_paths(
        save_filename,
        check_hash=True,
        redownload_on_hash_mismatch=True,
        url=mappings_url,
    )

    mapping_filenames = {
        "hashes": "hashes.json",
        "name_aliases": "name_mappings.json",
        "paths_and_uids": "uid-cache.json",
        "structure_names_by_uid": "structure-names-mapping-cache.json",
    }

    mapping_paths = {}
    for label, filename in mapping_filenames.items():
        mapping_path = [path for path in mappings_paths if path.name == filename]
        if len(mapping_path) != 1:
            raise ValueError(f"Expected exactly one file named {filename}.")

        mapping_paths[label] = mapping_path[0]

    hash_path = mapping_paths["hashes"]

    data_path_root = hash_path.parent.parent

    with open(mapping_paths["paths_and_uids"]) as f:
        uid_cache = json.load(f)

    (
        ct_image_paths,
        structure_set_paths,
        ct_uid_to_structure_uid,
        structure_uid_to_ct_uids,
    ) = indexing.read_uid_cache(data_path_root, uid_cache)

    with open(mapping_paths["structure_names_by_uid"]) as f:
        structure_names_cache = json.load(f)

    structure_names_by_ct_uid = structure_names_cache["structure_names_by_ct_uid"]
    structure_names_by_structure_set_uid = structure_names_cache[
        "structure_names_by_structure_set_uid"
    ]

    uid_to_url = {}

    for structure_uid, ct_uids in structure_uid_to_ct_uids.items():
        uid_to_url[structure_uid] = dicom_zip_url_pattern.format(
            dicom_type="RS", uid=structure_uid
        )

        for ct_uid in ct_uids:
            uid_to_url[ct_uid] = dicom_zip_url_pattern.format(
                dicom_type="CT", uid=ct_uid
            )

    names_map = filtering.load_names_mapping(mapping_paths["name_aliases"])

    return (
        data_path_root,
        structure_set_paths,
        ct_image_paths,
        ct_uid_to_structure_uid,
        structure_uid_to_ct_uids,
        names_map,
        structure_names_by_ct_uid,
        structure_names_by_structure_set_uid,
        uid_to_url,
        hash_path,
    )


def get_filtered_uids(filters, structure_uids=None):
    (
        _,
        _,
        _,
        ct_uid_to_structure_uid,
        structure_uid_to_ct_uids,
        _,
        structure_names_by_ct_uid,
        structure_names_by_structure_set_uid,
        _,
        _,
    ) = get_dataset_metadata()

    if structure_uids is None:
        structure_uids = structure_uid_to_ct_uids.keys()

    filtered_ct_uids = filtering.filter_ct_uids(
        structure_uids,
        structure_uid_to_ct_uids,
        structure_names_by_structure_set_uid,
        structure_names_by_ct_uid,
        **filters,
    )

    used_structure_uids = set()
    for ct_uid in filtered_ct_uids:
        used_structure_uids.add(ct_uid_to_structure_uid[ct_uid])

    used_structure_uids = list(used_structure_uids)
    sorted(used_structure_uids)

    return used_structure_uids, filtered_ct_uids


def create_dataset(ct_uids, structures_to_learn, expansion=5):
    (
        data_path_root,
        structure_set_paths,
        ct_image_paths,
        ct_uid_to_structure_uid,
        _,
        names_map,
        _,
        _,
        uid_to_url,
        hash_path,
    ) = get_dataset_metadata()

    def predownload_generator():
        for ct_uid in ct_uids:
            download_uid(data_path_root, ct_uid, uid_to_url, hash_path)

            structure_uid = ct_uid_to_structure_uid[ct_uid]
            download_uid(data_path_root, structure_uid, uid_to_url, hash_path)

            yield ct_uid

    pre_download_parameters = ((tf.string), (tf.TensorShape(())))
    pre_download_dataset = tf.data.Dataset.from_generator(
        predownload_generator, *pre_download_parameters
    )

    def generator():
        for ct_uid in pre_download_dataset.prefetch(30):
            ct_uid = ct_uid.numpy().decode()
            x_grid, y_grid, input_array, output_array = numpy_input_output_from_cache(
                data_path_root,
                structure_set_paths,
                ct_image_paths,
                ct_uid_to_structure_uid,
                names_map,
                ct_uid,
                structures_to_learn,
                expansion=expansion,
            )
            input_array = input_array[:, :, None]

            yield ct_uid, x_grid, y_grid, input_array, output_array

    parameters = (
        (tf.string, tf.float64, tf.float64, tf.int32, tf.float64),
        (
            tf.TensorShape(()),
            tf.TensorShape([512]),
            tf.TensorShape([512]),
            tf.TensorShape([512, 512, 1]),
            tf.TensorShape([512, 512, len(structures_to_learn)]),
        ),
    )

    dataset = tf.data.Dataset.from_generator(generator, *parameters)

    return dataset


def get_contours_by_ct_uid(structure_set, number_to_name_map):
    contours_by_ct_uid = {}

    for roi_contour_sequence_item in structure_set.ROIContourSequence:
        try:
            structure_name = number_to_name_map[
                roi_contour_sequence_item.ReferencedROINumber
            ]
        except KeyError:
            continue

        for contour_sequence_item in roi_contour_sequence_item.ContourSequence:
            ct_uid = contour_sequence_item.ContourImageSequence[
                0
            ].ReferencedSOPInstanceUID

            try:
                _ = contours_by_ct_uid[ct_uid]
            except KeyError:
                contours_by_ct_uid[ct_uid] = dict()

            try:
                contours_by_ct_uid[ct_uid][structure_name].append(
                    contour_sequence_item.ContourData
                )
            except KeyError:
                contours_by_ct_uid[ct_uid][structure_name] = [
                    contour_sequence_item.ContourData
                ]

    return contours_by_ct_uid


def get_filename_from_url(url):
    filename = url.split("/")[-1]

    return filename


def download_uid(data_path_root, uid, uid_to_url, hash_path):
    download_directory_name = data_path_root.name

    url = uid_to_url[uid]
    filename = get_filename_from_url(url)
    save_filepath = pathlib.Path(download_directory_name).joinpath("dicom", filename)

    paths = download.zip_data_paths(
        save_filepath,
        check_hash=True,
        redownload_on_hash_mismatch=True,
        delete_when_no_hash_found=True,
        url=url,
        hash_filepath=hash_path,
    )

    if len(paths) != 1:
        raise ValueError("Expected only 1 path to be downloaded.")

    return paths[0]


def create_input_ct_image(dcm_ct):
    x_grid, y_grid, _ = mask.get_grid(dcm_ct)

    return x_grid, y_grid, dcm_ct.pixel_array


def create_output_mask(dcm_ct, contours_by_ct_uid, structure, ct_uid, expansion=5):
    _, _, ct_size = mask.get_grid(dcm_ct)

    contours_on_this_slice = contours_by_ct_uid[ct_uid].keys()
    if structure in contours_on_this_slice:
        original_contours = contours_by_ct_uid[ct_uid][structure]
        _, _, calculated_mask = mask.calculate_anti_aliased_mask(
            original_contours, dcm_ct, expansion=expansion
        )
    else:
        calculated_mask = np.zeros(ct_size) - 1

    return calculated_mask


def numpy_input_output_from_cache(
    data_path_root,
    structure_set_paths,
    ct_image_paths,
    ct_uid_to_structure_uid,
    names_map,
    ct_uid,
    structures_to_learn,
    expansion=5,
):
    data_path_root = pathlib.Path(data_path_root)

    @functools.lru_cache()
    def get_dcm_ct_from_uid(ct_uid):
        ct_path = ct_image_paths[ct_uid]
        dcm_ct = pydicom.read_file(ct_path, force=True)

        dcm_ct.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian

        return dcm_ct

    @functools.lru_cache()
    def get_dcm_structure_from_uid(structure_set_uid):
        structure_set_path = structure_set_paths[structure_set_uid]

        dcm_structure = pydicom.read_file(
            structure_set_path,
            force=True,
            specific_tags=["ROIContourSequence", "StructureSetROISequence"],
        )

        return dcm_structure

    @functools.lru_cache()
    def get_contours_by_ct_uid_from_structure_uid(structure_set_uid):
        dcm_structure = get_dcm_structure_from_uid(structure_set_uid)

        number_to_name_map = {
            roi_sequence_item.ROINumber: names_map[roi_sequence_item.ROIName]
            for roi_sequence_item in dcm_structure.StructureSetROISequence
            if names_map[roi_sequence_item.ROIName] is not None
        }

        contours_by_ct_uid = get_contours_by_ct_uid(dcm_structure, number_to_name_map)

        return contours_by_ct_uid

    npz_directory = data_path_root.joinpath("npz_cache")
    npz_directory.mkdir(parents=True, exist_ok=True)

    npz_input_path = npz_directory.joinpath(f"{ct_uid}_input.npz")
    npz_output_paths = {
        structure: npz_directory.joinpath(
            f"{ct_uid}_output_{structure}_expansion_{expansion}.npz"
        )
        for structure in structures_to_learn
    }

    try:
        input_data = np.load(npz_input_path)
    except FileNotFoundError:
        dcm_ct = get_dcm_ct_from_uid(ct_uid)
        x_grid, y_grid, input_array = create_input_ct_image(dcm_ct)

        np.savez(npz_input_path, x_grid=x_grid, y_grid=y_grid, input_array=input_array)
    else:
        x_grid = input_data["x_grid"]
        y_grid = input_data["y_grid"]
        input_array = input_data["input_array"]

    output_array = np.ones((*np.shape(input_array), len(structures_to_learn)))
    for i, structure in enumerate(structures_to_learn):
        npz_path = npz_output_paths[structure]

        try:
            output_data = np.load(npz_path)
        except FileNotFoundError:
            structure_set_uid = ct_uid_to_structure_uid[ct_uid]
            contours_by_ct_uid = get_contours_by_ct_uid_from_structure_uid(
                structure_set_uid
            )

            dcm_ct = get_dcm_ct_from_uid(ct_uid)
            calculated_mask = create_output_mask(
                dcm_ct, contours_by_ct_uid, structure, ct_uid, expansion=expansion
            )

            np.savez(npz_path, mask=calculated_mask)
        else:
            calculated_mask = output_data["mask"]

        output_array[:, :, i] = calculated_mask

    return x_grid, y_grid, input_array, output_array



# Copyright (C) 2015, 2019 Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable = protected-access


"""Tests for gamma shell."""


import numpy as np

import pymedphys
import pymedphys._utilities.createshells


def does_gamma_scale_as_expected(
    init_distance_threshold, threshold_ratio, scales_to_test
):
    coords, reference, evaluation, _ = get_dummy_gamma_set()

    all_scales = np.concatenate([[1], scales_to_test])

    distance_thresholds_to_test = init_distance_threshold * all_scales
    dose_thresholds_to_test = distance_thresholds_to_test * threshold_ratio

    gamma_results = []
    for dose, distance in zip(dose_thresholds_to_test, distance_thresholds_to_test):
        gamma_results.append(
            pymedphys.gamma(
                coords,
                reference,
                coords,
                evaluation,
                dose,
                distance,
                lower_percent_dose_cutoff=0,
            )
        )

    for i, scale in enumerate(scales_to_test):
        print(scale)

        abs_diff = np.abs(gamma_results[i + 1] - gamma_results[0] / scale)

        print(np.max(abs_diff))

        assert np.all(abs_diff <= 0.1)


def test_a_set_of_gamma_scaling():
    does_gamma_scale_as_expected(0.6, 10, [2, 3, 4])
    does_gamma_scale_as_expected(0.6, 15, [2, 3, 4])
    does_gamma_scale_as_expected(0.6, 20, [2, 3, 4])


def test_multiple_threshold_inputs():
    coords, reference, evaluation, _ = get_dummy_gamma_set()

    pymedphys.gamma(
        coords,
        reference,
        coords,
        evaluation,
        [3],
        [0.3, 0.5],
        lower_percent_dose_cutoff=0,
    )


def test_lower_dose_threshold():
    """Verify that the lower dose threshold works as expected"""
    ref = [0, 1, 1.9, 2, 2.1, 3, 4, 5, 10, 10]
    coords_ref = (np.arange(len(ref)),)

    evl = [10] * (len(ref) + 2)
    coords_evl = (np.arange(len(evl)) - 4,)

    result = pymedphys.gamma(coords_ref, ref, coords_evl, evl, 10, 1)

    assert np.array_equal(ref < 0.2 * np.max(ref), np.isnan(result))


def get_dummy_gamma_set():
    grid_x = np.arange(0, 1, 0.1)
    grid_y = np.arange(0, 1.2, 0.1)
    grid_z = np.arange(0, 1.4, 0.1)
    dimensions = (len(grid_x), len(grid_y), len(grid_z))
    coords = (grid_x, grid_y, grid_z)

    reference = np.zeros(dimensions)
    reference[3:-2:, 4:-2:, 5:-2:] = 1.015

    evaluation = np.zeros(dimensions)
    evaluation[2:-2:, 2:-2:, 2:-2:] = 1

    expected_gamma = np.zeros(dimensions)
    expected_gamma[2:-2:, 2:-2:, 2:-2:] = 0.4
    expected_gamma[3:-3:, 3:-3:, 3:-3:] = 0.7
    expected_gamma[4:-4:, 4:-4:, 4:-4:] = 1
    expected_gamma[3:-2:, 4:-2:, 5:-2:] = 0.5

    return coords, reference, evaluation, expected_gamma


def test_regression_of_gamma_3d():
    """Test for changes in expected 3D gamma."""
    coords, reference, evaluation, expected_gamma = get_dummy_gamma_set()

    gamma3d = np.round(
        pymedphys.gamma(
            coords, reference, coords, evaluation, 3, 0.3, lower_percent_dose_cutoff=0
        ),
        decimals=1,
    )

    assert np.all(expected_gamma == gamma3d)


def test_regression_of_gamma_2d():
    """Test for changes in expected 2D gamma."""
    coords, reference, evaluation, expected_gamma = get_dummy_gamma_set()

    gamma2d = np.round(
        pymedphys.gamma(
            coords[1::],
            reference[5, :, :],
            coords[1::],
            evaluation[5, :, :],
            3,
            0.3,
            lower_percent_dose_cutoff=0,
        ),
        decimals=1,
    )

    assert np.all(expected_gamma[5, :, :] == gamma2d)


def test_regression_of_gamma_1d():
    """Test for changes in expected 3D gamma."""

    coords, reference, evaluation, expected_gamma = get_dummy_gamma_set()

    gamma1d = np.round(
        pymedphys.gamma(
            coords[2],
            reference[5, 5, :],
            coords[2],
            evaluation[5, 5, :],
            3,
            0.3,
            lower_percent_dose_cutoff=0,
        ),
        decimals=1,
    )

    assert np.all(expected_gamma[5, 5, :] == gamma1d)


def test_coords_stepsize():
    """Testing correct stepsize implementation.

    Confirm that the the largest distance between one point and any other
    is less than the defined step size
    """
    distance_step_size = 0.1
    num_dimensions = 3
    distance = 1

    x, y, z = pymedphys._utilities.createshells.calculate_coordinates_shell(
        distance, num_dimensions, distance_step_size
    )

    distance_between_coords = np.sqrt(
        (x[:, None] - x[None, :]) ** 2
        + (y[:, None] - y[None, :]) ** 2
        + (z[:, None] - z[None, :]) ** 2
    )

    distance_between_coords[distance_between_coords == 0] = np.nan

    largest_difference = np.max(np.nanmin(distance_between_coords, axis=0))

    assert largest_difference <= distance_step_size
    assert largest_difference > distance_step_size * 0.9


def test_distance_0_gives_1_point():
    """Testing correct stepsize implementation.

    Confirm that the the largest distance between one point and any other
    is less than the defined step size
    """
    distance_step_size = 0.1
    num_dimensions = 3
    distance = 0

    x, y, z = pymedphys._utilities.createshells.calculate_coordinates_shell(
        distance, num_dimensions, distance_step_size
    )

    assert len(x) == 1 & len(y) == 1 & len(z) == 1



# Copyright (C) 2020 Stuart Swerdloff, Simon Biggs
# Copyright (C) 2018 Matthew Jennings, Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .api import anonymise_cli, anonymise_dataset, anonymise_directory, anonymise_file
from .core import (
    IDENTIFYING_KEYWORDS_FILEPATH,
    get_baseline_keyword_vr_dict,
    get_default_identifying_keywords,
    is_anonymised_dataset,
    is_anonymised_directory,
    is_anonymised_file,
    label_dicom_filepath_as_anonymised,
)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy

from pymedphys._imports import numpy as np


def add_data_to_control_point(template, data, i):
    cp = deepcopy(template)
    cp.ControlPointIndex = str(i)

    cp.GantryAngle = data["gantry_angle"][i]
    cp.GantryRotationDirection = data["gantry_movement"][i]

    cp.BeamLimitingDeviceAngle = data["collimator_angle"][i]
    cp.BeamLimitingDeviceRotationDirection = data["collimator_movement"][i]

    collimation = cp.BeamLimitingDevicePositionSequence
    collimation[0].LeafJawPositions = data["jaw"][i]
    collimation[1].LeafJawPositions = data["mlc"][i]

    cp.CumulativeMetersetWeight = np.around(
        data["monitor_units"][i] / data["monitor_units"][-1], decimals=6
    )

    return cp


def build_control_points(initial_cp_template, subsequent_cp_template, data):
    number_of_control_points = len(data["monitor_units"])

    cps = []
    for i in range(number_of_control_points):
        if i == 0:
            template = initial_cp_template
        else:
            template = subsequent_cp_template

        cps.append(add_data_to_control_point(template, data, i))

    return cps


def replace_fraction_group(
    created_dicom, beam_meterset, beam_index, fraction_group_index
):
    fraction_group = created_dicom.FractionGroupSequence[fraction_group_index]
    referenced_beam = fraction_group.ReferencedBeamSequence[beam_index]
    referenced_beam.BeamMeterset = str(beam_meterset)
    fraction_group.ReferencedBeamSequence = [referenced_beam]
    created_dicom.FractionGroupSequence = [fraction_group]


def replace_beam_sequence(created_dicom, all_control_points, beam_index):
    beam = created_dicom.BeamSequence[beam_index]
    beam.ControlPointSequence = all_control_points
    beam.NumberOfControlPoints = len(all_control_points)
    created_dicom.BeamSequence = [beam]


def restore_trailing_zeros(created_dicom):
    for beam_sequence in created_dicom.BeamSequence:
        for control_point in beam_sequence.ControlPointSequence:
            current_value = float(control_point.CumulativeMetersetWeight)
            new_value = "{0:.6f}".format(current_value)

            control_point.CumulativeMetersetWeight = new_value


def merge_beam_sequences(dicoms_by_gantry_angle):
    merged = dicoms_by_gantry_angle[0]

    for dicom in dicoms_by_gantry_angle[1::]:
        merged.BeamSequence.append(dicom.BeamSequence[0])
        merged.FractionGroupSequence[0].ReferencedBeamSequence.append(
            dicom.FractionGroupSequence[0].ReferencedBeamSequence[0]
        )

    return merged



import copy
import functools
import json
import logging
import os
import subprocess
import tempfile
from copy import deepcopy
from os.path import basename, dirname, exists
from os.path import join as pjoin
from shutil import copyfile
from uuid import uuid4

import pytest

import pydicom
import pydicom.datadict
import pydicom.dataset
import pydicom.filereader
import pydicom.tag

from pymedphys._data import download
from pymedphys._dicom import create
from pymedphys._dicom.anonymise import (
    IDENTIFYING_KEYWORDS_FILEPATH,
    anonymise_directory,
    anonymise_file,
    get_baseline_keyword_vr_dict,
    get_default_identifying_keywords,
    is_anonymised_dataset,
    is_anonymised_directory,
    is_anonymised_file,
    label_dicom_filepath_as_anonymised,
)
from pymedphys._dicom.constants import get_baseline_dicom_dict
from pymedphys._dicom.create import dicom_dataset_from_dict
from pymedphys._dicom.utilities import remove_file
from pymedphys.dicom import anonymise as anonymise_dataset

# TODO: TEST_ANON_BASENAME will probably instead need to contain the
# PYMEDPHYS_ROOT_UID (or similar) when anonymisation of UIDS is
# implemented
TEST_ANON_BASENAME = (
    "RP.1.2.246.352.71.5.53598612033.430805.20190416135558_Anonymised.dcm"
)

TEST_ANON_BASENAME_DICT = {
    "RP.almost_anonymised.dcm": "RP.1.2.246.352.71.5.53598612033.430805.20190416135558_Anonymised.dcm",
    "RIBT.not_quite_anonymised.dcm": "RIBT.1.2.392.200036.9123.100.30.310.200.12.1.20191125110540243000_Anonymised.dcm",
}

VR_NON_ANONYMOUS_REPLACEMENT_VALUE_DICT = {
    "AE": "AnAETitle",
    "AS": "1Y",
    "CS": "SMITH",
    "DA": "20190429",
    "DS": "11111111.9",
    "DT": "20190429000700.000000",
    "LO": "Smith",
    "LT": "LongText",
    "OB": (2).to_bytes(2, "little"),
    "OB or OW": (2).to_bytes(2, "little"),
    "OW": (2).to_bytes(2, "little"),
    "PN": "Smith",
    "SH": "Smith",
    "SQ": [pydicom.dataset.Dataset(), pydicom.dataset.Dataset()],
    "ST": "Smith",
    "TM": "000700.000000",
    "UI": "1118",
    "US": 11111,
}


@functools.lru_cache()
def get_rtplan_test_file_path():
    data_paths = _download_rtplan_test_file()
    test_rtplan_path = next(
        x for x in data_paths if x.name == "RP.almost_anonymised.dcm"
    )
    test_rtplan_file_path = str(test_rtplan_path.absolute())
    return test_rtplan_file_path


@functools.lru_cache()
def _download_rtplan_test_file():
    data_paths = download.zip_data_paths("rtplan-anonymisation.zip")
    return data_paths


@functools.lru_cache()
def _download_treatment_record_test_file():
    data_paths = download.zip_data_paths("treatment-record-anonymisation.zip")
    return data_paths


@functools.lru_cache()
def get_treatment_record_test_file_path():
    """
    Downloads the required test data from zenodo and returns a string representing
    the absolute path including the test filename, appropriate to the platform/file system.
    Because this function is wrapped with lru_cache, concurrently revised contents on zenodo will not
    get pulled down while the python instance is running (the tests).
    If you update the content on zenodo, re-run the tests.

    Parameters
    ----------
    None

    Returns
    -------
    string:representing the absolute path including the test filename, appropriate to the platform/file system.

    """
    data_paths = _download_treatment_record_test_file()
    test_treatment_record_path = next(
        x for x in data_paths if x.name == "RIBT.not_quite_anonymised.dcm"
    )
    test_treatment_record_file_path = str(test_treatment_record_path.absolute())
    return test_treatment_record_file_path


def get_test_filepaths():
    return [get_rtplan_test_file_path(), get_treatment_record_test_file_path()]


def _check_is_anonymised_dataset_file_and_dir(
    ds, tmp_path, test_file_path, anon_is_expected=True, ignore_private_tags=False
):
    temp_filepath = str(tmp_path / "test.dcm")

    try:
        create.set_default_transfer_syntax(ds)

        ds.file_meta = pydicom.filereader.read_file_meta_info(test_file_path)

        ds.save_as(temp_filepath, write_like_original=False)

        if anon_is_expected:
            assert is_anonymised_dataset(ds, ignore_private_tags)
            assert is_anonymised_file(temp_filepath, ignore_private_tags)
            assert is_anonymised_directory(tmp_path, ignore_private_tags)
        else:
            assert not is_anonymised_dataset(ds, ignore_private_tags)
            assert not is_anonymised_file(temp_filepath, ignore_private_tags)
            assert not is_anonymised_directory(tmp_path, ignore_private_tags)
    finally:
        remove_file(temp_filepath)


def _get_non_anonymous_replacement_value(keyword):
    """Get an appropriate dummy non-anonymised value for a DICOM element based
    on its value representation (VR)"""
    vr = get_baseline_keyword_vr_dict()[keyword]
    return VR_NON_ANONYMOUS_REPLACEMENT_VALUE_DICT[vr]


@pytest.mark.pydicom
def test_anonymised_dataset_with_nested_name():
    nested_name = dicom_dataset_from_dict(
        {"OverrideSequence": [{"OperatorsName": "George"}]}
    )

    assert not is_anonymised_dataset(nested_name)


@pytest.mark.pydicom
def test_anonymised_dataset_with_empty_patient_sex():
    blank_sex_ds = dicom_dataset_from_dict({"PatientSex": None})
    assert is_anonymised_dataset(blank_sex_ds)
    hardcode_replace_ds = anonymise_dataset(blank_sex_ds)
    assert hardcode_replace_ds["PatientSex"].value is None


@pytest.mark.slow
@pytest.mark.pydicom
def test_alternative_identifying_keywords():
    alternative_keyword_list = copy.copy(get_default_identifying_keywords())
    alternative_keyword_list.append("SOPInstanceUID")
    test_file_path = get_treatment_record_test_file_path()
    ds_test = pydicom.dcmread(test_file_path, force=True)
    with tempfile.TemporaryDirectory() as output_directory:
        anon_private_filepath = anonymise_file(
            test_file_path,
            output_filepath=output_directory,
            delete_private_tags=True,
            identifying_keywords=alternative_keyword_list,
        )
        ds_anon = pydicom.dcmread(anon_private_filepath, force=True)

        assert is_anonymised_file(anon_private_filepath, ignore_private_tags=False)
        assert ds_test["SOPInstanceUID"].value != ds_anon["SOPInstanceUID"].value


@pytest.mark.slow
@pytest.mark.pydicom
def test_anonymise_dataset_and_all_is_anonymised_functions(tmp_path):

    # Create dataset with one instance of every identifying keyword and
    # run basic anonymisation tests
    test_file_path = get_rtplan_test_file_path()
    ds = pydicom.dataset.Dataset()
    for keyword in get_default_identifying_keywords():
        # Ignore file meta elements for now
        tag = hex(pydicom.datadict.tag_for_keyword(keyword))
        if pydicom.tag.Tag(tag).group == 0x0002:
            continue

        value = _get_non_anonymous_replacement_value(keyword)
        setattr(ds, keyword, value)

    _check_is_anonymised_dataset_file_and_dir(
        ds, tmp_path, test_file_path, anon_is_expected=False
    )

    ds_anon = anonymise_dataset(ds)
    _check_is_anonymised_dataset_file_and_dir(
        ds_anon, tmp_path, test_file_path, anon_is_expected=True
    )

    # Test the anonymisation and check functions for each identifying
    # element individually.
    for elem in ds_anon.iterall():

        # TODO: AffectedSOPInstanceUID and RequestedSOPInstanceUID
        # are not writing to file. Investigate when UID anonymisation is
        # implemented.
        if elem.keyword in ("AffectedSOPInstanceUID", "RequestedSOPInstanceUID"):
            continue

        ds_single_non_anon_value = deepcopy(ds_anon)
        setattr(
            ds_single_non_anon_value,
            elem.keyword,
            _get_non_anonymous_replacement_value(elem.keyword),
        )
        _check_is_anonymised_dataset_file_and_dir(
            ds_single_non_anon_value, tmp_path, test_file_path, anon_is_expected=False
        )
        ds_single_anon = anonymise_dataset(ds_single_non_anon_value)
        _check_is_anonymised_dataset_file_and_dir(
            ds_single_anon, tmp_path, test_file_path, anon_is_expected=True
        )

    # Test correct handling of private tags
    ds_anon.add(pydicom.dataset.DataElement(0x0043102B, "SS", [4, 4, 0, 0]))
    _check_is_anonymised_dataset_file_and_dir(
        ds_anon,
        tmp_path,
        test_file_path,
        anon_is_expected=False,
        ignore_private_tags=False,
    )
    _check_is_anonymised_dataset_file_and_dir(
        ds_anon,
        tmp_path,
        test_file_path,
        anon_is_expected=True,
        ignore_private_tags=True,
    )

    ds_anon.remove_private_tags()
    _check_is_anonymised_dataset_file_and_dir(
        ds_anon,
        tmp_path,
        test_file_path,
        anon_is_expected=True,
        ignore_private_tags=False,
    )

    # Test blank anonymisation
    # # Sanity check
    _check_is_anonymised_dataset_file_and_dir(
        ds, tmp_path, test_file_path, anon_is_expected=False
    )

    ds_anon_blank = anonymise_dataset(ds, replace_values=False)
    _check_is_anonymised_dataset_file_and_dir(
        ds_anon_blank, tmp_path, test_file_path, anon_is_expected=True
    )

    # Test handling of unknown tags by removing PatientName from
    # baseline dict
    patient_name_tag = pydicom.datadict.tag_for_keyword("PatientName")

    try:
        patient_name = get_baseline_dicom_dict().pop(patient_name_tag)

        with pytest.raises(ValueError) as e_info:
            anonymise_dataset(ds)
        assert str(e_info.value).count(
            "At least one of the non-private tags "
            "within your DICOM file is not within "
            "PyMedPhys's copy of the DICOM dictionary."
        )

        ds_anon_delete_unknown = anonymise_dataset(ds, delete_unknown_tags=True)
        _check_is_anonymised_dataset_file_and_dir(
            ds_anon_delete_unknown, tmp_path, test_file_path, anon_is_expected=True
        )
        with pytest.raises(AttributeError) as e_info:
            ds_anon_delete_unknown.PatientName  # pylint: disable = pointless-statement
        assert str(e_info.value).count(
            "'Dataset' object has no attribute " "'PatientName'"
        )

        ds_anon_ignore_unknown = anonymise_dataset(ds, delete_unknown_tags=False)
        _check_is_anonymised_dataset_file_and_dir(
            ds_anon_ignore_unknown, tmp_path, test_file_path, anon_is_expected=True
        )
        assert patient_name_tag in ds_anon_ignore_unknown

    finally:
        get_baseline_dicom_dict().setdefault(patient_name_tag, patient_name)

    # Test copy_dataset=False:
    anonymise_dataset(ds, copy_dataset=False)
    assert is_anonymised_dataset(ds)


@pytest.mark.pydicom
def test_anonymise_file():
    for test_file_path in get_test_filepaths():
        _test_anonymise_file_at_path(test_file_path)


def _test_anonymise_file_at_path(test_file_path):
    assert not is_anonymised_file(test_file_path)
    temp_basename = "{}_{}.dcm".format(
        ".".join(test_file_path.split(".")[:-1]), uuid4()
    )

    temp_filepath = pjoin(dirname(test_file_path), temp_basename)
    anon_private_filepath = ""
    anon_filepath_orig = ""
    anon_filepath_pres = ""

    try:
        # Private tag handling
        anon_private_filepath = anonymise_file(
            test_file_path, delete_private_tags=False
        )
        assert not is_anonymised_file(anon_private_filepath, ignore_private_tags=False)
        assert is_anonymised_file(anon_private_filepath, ignore_private_tags=True)

        anon_private_filepath = anonymise_file(test_file_path, delete_private_tags=True)
        assert is_anonymised_file(anon_private_filepath, ignore_private_tags=False)

        # Filename is anonymised?
        assert (
            basename(anon_private_filepath)
            == TEST_ANON_BASENAME_DICT[basename(test_file_path)]
        )

        # Deletion of original file
        copyfile(test_file_path, temp_filepath)

        anon_filepath_orig = anonymise_file(temp_filepath, delete_original_file=True)
        assert is_anonymised_file(anon_filepath_orig)
        assert not exists(temp_filepath)

        # Preservation of filename if desired
        expected_filepath = "{}_Anonymised.dcm".format(
            ".".join(test_file_path.split(".")[:-1])
        )
        anon_filepath_pres = anonymise_file(test_file_path, anonymise_filename=False)
        assert anon_filepath_pres == expected_filepath

    finally:
        remove_file(temp_filepath)
        remove_file(anon_private_filepath)
        remove_file(anon_filepath_orig)
        remove_file(anon_filepath_pres)


@pytest.mark.pydicom
def test_anonymise_directory(tmp_path):
    temp_filepath = tmp_path / "test.dcm"
    temp_anon_filepath = label_dicom_filepath_as_anonymised(temp_filepath)

    temp_record_filepath = tmp_path / "test_record.dcm"
    temp_anon_record_filepath = label_dicom_filepath_as_anonymised(temp_record_filepath)

    try:
        copyfile(get_rtplan_test_file_path(), temp_filepath)
        assert not is_anonymised_directory(tmp_path)

        # Test file deletion
        anon_path_list = anonymise_directory(
            tmp_path, delete_original_files=False, anonymise_filenames=False
        )
        # # File should be anonymised but not dir, since original file
        # # is still present.
        assert is_anonymised_file(temp_anon_filepath)
        assert exists(temp_filepath)
        assert not is_anonymised_directory(tmp_path)
        assert anon_path_list is not None
        assert anon_path_list[0] == temp_anon_filepath

        remove_file(temp_anon_filepath)
        anon_path_list = anonymise_directory(
            tmp_path, delete_original_files=True, anonymise_filenames=False
        )
        # # File and dir should be anonymised since original file should
        # # have been deleted.
        assert is_anonymised_file(temp_anon_filepath)
        assert not exists(temp_filepath)
        assert is_anonymised_directory(tmp_path)
        assert anon_path_list[0] == temp_anon_filepath

        # Test fail fast vs. fail at last
        # if the function fails fast, the specified removal
        # will not take place
        # if the function does not fail fail, the specified
        # removal will take place
        logging.warning("Testing fail fast")
        remove_file(temp_anon_filepath)
        copyfile(get_rtplan_test_file_path(), temp_filepath)
        copyfile(get_treatment_record_test_file_path(), temp_record_filepath)
        ds_record = pydicom.dcmread(temp_record_filepath, force=True)
        # deliberately add a DICOM element that is not in the current
        # dictionary
        ds_record.add_new([0x300A, 0x9999], "FL", [1.0, 1.0])
        pydicom.dcmwrite(temp_record_filepath, ds_record)
        with pytest.raises((KeyError, ValueError)):
            anon_path_list = anonymise_directory(
                tmp_path,
                delete_original_files=True,
                anonymise_filenames=False,
                fail_fast=True,
            )
            logging.warning(anon_path_list)
            assert exists(temp_filepath)

        with pytest.raises((KeyError, ValueError)):
            anon_path_list = anonymise_directory(
                tmp_path,
                delete_original_files=True,
                anonymise_filenames=False,
                fail_fast=False,
            )
            logging.warning(anon_path_list)
            assert not exists(temp_filepath)

    finally:
        remove_file(temp_anon_filepath)
        remove_file(temp_anon_record_filepath)


@pytest.mark.slow
@pytest.mark.pydicom
@pytest.mark.skipif(
    "SUBPACKAGE" in os.environ, reason="Need to extract CLI out of subpackages"
)
def test_anonymise_cli(tmp_path):
    for test_file_path in get_test_filepaths():
        _test_anonymise_cli_for_file(tmp_path, test_file_path)


def _test_anonymise_cli_for_file(tmp_path, test_file_path):
    temp_filepath = str(tmp_path / "test.dcm")
    try:
        logging.info("CLI test on %s", test_file_path)

        copyfile(test_file_path, temp_filepath)
        test_anon_basename = TEST_ANON_BASENAME_DICT[basename(test_file_path)]
        temp_anon_filepath = str(tmp_path / test_anon_basename)
        # Basic file anonymisation
        assert not is_anonymised_file(temp_filepath)
        assert not exists(temp_anon_filepath)

        anon_file_command = "pymedphys dicom anonymise".split() + [temp_filepath]
        try:
            subprocess.check_call(anon_file_command)
            assert is_anonymised_file(temp_anon_filepath)
            assert exists(temp_filepath)
        finally:
            remove_file(temp_anon_filepath)

        # File anonymisation - preserve filenames
        assert not is_anonymised_file(temp_filepath)

        expected_anon_filepath = label_dicom_filepath_as_anonymised(temp_filepath)
        assert not exists(expected_anon_filepath)

        anon_file_pres_command = "pymedphys dicom anonymise -f".split() + [
            temp_filepath
        ]
        try:
            subprocess.check_call(anon_file_pres_command)
            assert is_anonymised_file(expected_anon_filepath)
            assert exists(temp_filepath)
        finally:
            remove_file(expected_anon_filepath)

        # File anonymisation - clear values
        assert not is_anonymised_file(temp_filepath)
        assert not exists(temp_anon_filepath)

        temp_cleared_anon_filepath = str(tmp_path / test_anon_basename)

        anon_file_clear_command = "pymedphys dicom anonymise -c".split() + [
            temp_filepath
        ]
        try:
            subprocess.check_call(anon_file_clear_command)
            assert is_anonymised_file(temp_cleared_anon_filepath)
            assert pydicom.dcmread(temp_cleared_anon_filepath).PatientName == ""
            assert exists(temp_filepath)
        finally:
            remove_file(temp_cleared_anon_filepath)

        # File anonymisation - leave keywords unchanged
        assert not is_anonymised_file(temp_filepath)
        assert not exists(temp_anon_filepath)

        anon_file_keep_command = (
            "pymedphys dicom anonymise".split()
            + [temp_filepath]
            + "-k PatientName".split()
        )
        try:
            subprocess.check_call(anon_file_keep_command)
            assert not is_anonymised_file(temp_anon_filepath)
            ds = pydicom.dcmread(temp_anon_filepath)
            ds.PatientName = "Anonymous"
            assert is_anonymised_dataset(ds)
            assert exists(temp_filepath)
        finally:
            remove_file(temp_anon_filepath)

        # File anonymisation - private tag handling
        assert not is_anonymised_file(temp_filepath)
        assert not exists(temp_anon_filepath)

        anon_file_private_command = "pymedphys dicom anonymise -p".split() + [
            temp_filepath
        ]
        try:
            subprocess.check_call(anon_file_private_command)
            assert not is_anonymised_file(temp_anon_filepath)
            assert is_anonymised_file(temp_anon_filepath, ignore_private_tags=True)
            assert exists(temp_filepath)
        finally:
            remove_file(temp_anon_filepath)

        # TODO: File anonymisation - unknown tag handling
        # # Calling a subprocess reloads BASELINE_DICOM_DICT...

        # Basic dir anonymisation
        assert not is_anonymised_directory(tmp_path)
        assert not exists(temp_anon_filepath)

        anon_dir_command = "pymedphys dicom anonymise".split() + [str(tmp_path)]
        try:
            subprocess.check_call(anon_dir_command)
            assert is_anonymised_file(temp_anon_filepath)
            assert exists(temp_filepath)
        finally:
            remove_file(temp_anon_filepath)
    finally:
        remove_file(temp_filepath)


@pytest.mark.pydicom
def test_tags_to_anonymise_in_dicom_dict_baseline(save_new_identifying_keywords=False):
    baseline_keywords = [val[4] for val in get_baseline_dicom_dict().values()]
    assert set(get_default_identifying_keywords()).issubset(baseline_keywords)

    if save_new_identifying_keywords:
        with open(IDENTIFYING_KEYWORDS_FILEPATH, "w") as outfile:
            json.dump(
                get_default_identifying_keywords(), outfile, indent=2, sort_keys=True
            )

        # TODO: Keywords to add if/when anonymisation of UIDs is implemented:
        # "AffectedSOPInstanceUID",
        # "ConcatenationUID",
        # "ContextGroupExtensionCreatorUID",
        # "CreatorVersionUID",
        # "DeviceUID",
        # "DigitalSignatureUID",
        # "DimensionOrganizationUID",
        # "DoseReferenceUID",
        # "FailedSOPInstanceUIDList",
        # "FiducialUID",
        # "FrameOfReferenceUID",
        # "InstanceCreatorUID",
        # "IrradiationEventUID",
        # "LargePaletteColorLookupTableUID",
        # "MediaStorageSOPInstanceUID",
        # "PaletteColorLookupTableUID",
        # "ReferencedFrameOfReferenceUID",
        # "ReferencedGeneralPurposeScheduledProcedureStepTransactionUID",
        # "ReferencedSOPInstanceUID",
        # "ReferencedSOPInstanceUIDInFile",
        # "RelatedFrameOfReferenceUID",
        # "RequestedSOPInstanceUID",
        # "SeriesInstanceUID",
        # "SOPInstanceUID",
        # "StorageMediaFileSetUID",
        # "StudyInstanceUID",
        # "SynchronizationFrameOfReferenceUID",
        # "TemplateExtensionCreatorUID",
        # "TemplateExtensionOrganizationUID",
        # "TransactionUID",
        # "UID",



# Copyright (C) 2019 Matthew Jennings
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from pymedphys._utilities.transforms.affine import (
    rotate_about_vector,
    rotate_about_x,
    rotate_about_y,
    rotate_about_z,
    translate,
)


def get_single_and_multi_test_coords():

    x_single = 1
    y_single = 2
    z_single = 3
    test_coords_single = np.array((x_single, y_single, z_single))

    number_of_coordinates = 11
    x_multi = np.array(range(-1, number_of_coordinates - 1, 1))
    y_multi = np.array(range(-2, 2 * (number_of_coordinates - 1), 2))
    z_multi = np.array(range(-3, 3 * (number_of_coordinates - 1), 3))
    test_coords_multi = np.array((x_multi, y_multi, z_multi))

    return test_coords_single, test_coords_multi


def test_rotate_about_vector():

    test_coords_single, test_coords_multi = get_single_and_multi_test_coords()

    rotation_vector = np.array((2, 2, 2))

    # Rotation of coords about own axis yields no change
    assert np.allclose(
        rotate_about_vector(np.array((1, 1, 1)), rotation_vector, 52),
        np.array((1, 1, 1)),
    )
    assert np.allclose(
        rotate_about_vector(np.array((1, 1, 1)), rotation_vector, 52, active=True),
        np.array((1, 1, 1)),
    )

    # Rotation of coords about each cardinal axis yields same result as axis rotation functions
    # Singles
    assert np.allclose(
        rotate_about_vector(test_coords_single, np.array((1, 0, 0)), 27),
        rotate_about_x(test_coords_single, 27),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_single, np.array((1, 0, 0)), 27, active=True),
        rotate_about_x(test_coords_single, 27, active=True),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_single, np.array((0, 1, 0)), 256),
        rotate_about_y(test_coords_single, 256),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_single, np.array((0, 1, 0)), 256, active=True),
        rotate_about_y(test_coords_single, 256, active=True),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_single, np.array((0, 0, 1)), 182),
        rotate_about_z(test_coords_single, 182),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_single, np.array((0, 0, 1)), 182, active=True),
        rotate_about_z(test_coords_single, 182, active=True),
    )
    # Multis
    assert np.allclose(
        rotate_about_vector(test_coords_multi, np.array((1, 0, 0)), 27),
        rotate_about_x(test_coords_multi, 27),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_multi, np.array((1, 0, 0)), 27, active=True),
        rotate_about_x(test_coords_multi, 27, active=True),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_multi, np.array((0, 1, 0)), 256),
        rotate_about_y(test_coords_multi, 256),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_multi, np.array((0, 1, 0)), 256, active=True),
        rotate_about_y(test_coords_multi, 256, active=True),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_multi, np.array((0, 0, 1)), 182),
        rotate_about_z(test_coords_multi, 182),
    )
    assert np.allclose(
        rotate_about_vector(test_coords_multi, np.array((0, 0, 1)), 182, active=True),
        rotate_about_z(test_coords_multi, 182, active=True),
    )

    # Random test case verified against: http://www.nh.cas.cz/people/lazar/celler/online_tools.php
    assert np.allclose(
        rotate_about_vector(np.array((13, 29, -17)), np.array((41, -97, 61)), 197),
        np.array((-30.683355, 18.344803, -4.582567)),
    )
    assert np.allclose(
        rotate_about_vector(
            np.array((13, 29, -17)), np.array((41, -97, 61)), 197, active=True
        ),
        np.array((-30.106782, 11.185686, -16.354268)),
    )


def test_rotate_about_x():

    test_coords_single, test_coords_multi = get_single_and_multi_test_coords()

    x_s = test_coords_single[0]
    y_s = test_coords_single[1]
    z_s = test_coords_single[2]
    x_m = test_coords_multi[0]
    y_m = test_coords_multi[1]
    z_m = test_coords_multi[2]

    # Passive rotations
    assert np.allclose(
        rotate_about_x(test_coords_single, 90), np.array((x_s, z_s, np.negative(y_s)))
    )
    assert np.allclose(
        rotate_about_x(test_coords_single, 180),
        np.array((x_s, np.negative(y_s), np.negative(z_s))),
    )
    assert np.allclose(
        rotate_about_x(test_coords_single, 270), np.array((x_s, np.negative(z_s), y_s))
    )

    assert np.allclose(
        rotate_about_x(test_coords_multi, 90), np.array((x_m, z_m, np.negative(y_m)))
    )
    assert np.allclose(
        rotate_about_x(test_coords_multi, 180),
        np.array((x_m, np.negative(y_m), np.negative(z_m))),
    )
    assert np.allclose(
        rotate_about_x(test_coords_multi, 270), np.array((x_m, np.negative(z_m), y_m))
    )

    # Active rotations
    assert np.allclose(
        rotate_about_x(test_coords_single, 90, active=True),
        np.array((x_s, np.negative(z_s), y_s)),
    )
    assert np.allclose(
        rotate_about_x(test_coords_single, 180, active=True),
        np.array((x_s, np.negative(y_s), np.negative(z_s))),
    )
    assert np.allclose(
        rotate_about_x(test_coords_single, 270, active=True),
        np.array((x_s, z_s, np.negative(y_s))),
    )

    assert np.allclose(
        rotate_about_x(test_coords_multi, 90, active=True),
        np.array((x_m, np.negative(z_m), y_m)),
    )
    assert np.allclose(
        rotate_about_x(test_coords_multi, 180, active=True),
        np.array((x_m, np.negative(y_m), np.negative(z_m))),
    )
    assert np.allclose(
        rotate_about_x(test_coords_multi, 270, active=True),
        np.array((x_m, z_m, np.negative(y_m))),
    )


def test_rotate_about_y():

    test_coords_single, test_coords_multi = get_single_and_multi_test_coords()
    x_s = test_coords_single[0]
    y_s = test_coords_single[1]
    z_s = test_coords_single[2]
    x_m = test_coords_multi[0]
    y_m = test_coords_multi[1]
    z_m = test_coords_multi[2]

    # Passive
    assert np.allclose(
        rotate_about_y(test_coords_single, 90), np.array((np.negative(z_s), y_s, x_s))
    )
    assert np.allclose(
        rotate_about_y(test_coords_single, 180),
        np.array((np.negative(x_s), y_s, np.negative(z_s))),
    )
    assert np.allclose(
        rotate_about_y(test_coords_single, 270), np.array((z_s, y_s, np.negative(x_s)))
    )

    assert np.allclose(
        rotate_about_y(test_coords_multi, 90), np.array((np.negative(z_m), y_m, x_m))
    )
    assert np.allclose(
        rotate_about_y(test_coords_multi, 180),
        np.array((np.negative(x_m), y_m, np.negative(z_m))),
    )
    assert np.allclose(
        rotate_about_y(test_coords_multi, 270), np.array((z_m, y_m, np.negative(x_m)))
    )

    # Active
    assert np.allclose(
        rotate_about_y(test_coords_single, 90, active=True),
        np.array((z_s, y_s, np.negative(x_s))),
    )
    assert np.allclose(
        rotate_about_y(test_coords_single, 180, active=True),
        np.array((np.negative(x_s), y_s, np.negative(z_s))),
    )
    assert np.allclose(
        rotate_about_y(test_coords_single, 270, active=True),
        np.array((np.negative(z_s), y_s, x_s)),
    )

    assert np.allclose(
        rotate_about_y(test_coords_multi, 90, active=True),
        np.array((z_m, y_m, np.negative(x_m))),
    )
    assert np.allclose(
        rotate_about_y(test_coords_multi, 180, active=True),
        np.array((np.negative(x_m), y_m, np.negative(z_m))),
    )
    assert np.allclose(
        rotate_about_y(test_coords_multi, 270, active=True),
        np.array((np.negative(z_m), y_m, x_m)),
    )


def test_rotate_about_z():

    test_coords_single, test_coords_multi = get_single_and_multi_test_coords()

    x_s = test_coords_single[0]
    y_s = test_coords_single[1]
    z_s = test_coords_single[2]
    x_m = test_coords_multi[0]
    y_m = test_coords_multi[1]
    z_m = test_coords_multi[2]

    # Passive
    assert np.allclose(
        rotate_about_z(test_coords_single, 90), np.array((y_s, np.negative(x_s), z_s))
    )
    assert np.allclose(
        rotate_about_z(test_coords_single, 180),
        np.array((np.negative(x_s), np.negative(y_s), z_s)),
    )
    assert np.allclose(
        rotate_about_z(test_coords_single, 270), np.array((np.negative(y_s), x_s, z_s))
    )

    assert np.allclose(
        rotate_about_z(test_coords_multi, 90), np.array((y_m, np.negative(x_m), z_m))
    )
    assert np.allclose(
        rotate_about_z(test_coords_multi, 180),
        np.array((np.negative(x_m), np.negative(y_m), z_m)),
    )
    assert np.allclose(
        rotate_about_z(test_coords_multi, 270), np.array((np.negative(y_m), x_m, z_m))
    )

    # Active
    assert np.allclose(
        rotate_about_z(test_coords_single, 90, active=True),
        np.array((np.negative(y_s), x_s, z_s)),
    )
    assert np.allclose(
        rotate_about_z(test_coords_single, 180, active=True),
        np.array((np.negative(x_s), np.negative(y_s), z_s)),
    )
    assert np.allclose(
        rotate_about_z(test_coords_single, 270, active=True),
        np.array((y_s, np.negative(x_s), z_s)),
    )

    assert np.allclose(
        rotate_about_z(test_coords_multi, 90, active=True),
        np.array((np.negative(y_m), x_m, z_m)),
    )
    assert np.allclose(
        rotate_about_z(test_coords_multi, 180, active=True),
        np.array((np.negative(x_m), np.negative(y_m), z_m)),
    )
    assert np.allclose(
        rotate_about_z(test_coords_multi, 270, active=True),
        np.array((y_m, np.negative(x_m), z_m)),
    )


def test_translate():

    test_coords_single, test_coords_multi = get_single_and_multi_test_coords()

    assert np.allclose(
        translate(test_coords_single, np.array((10, 20, 30))), np.array((-9, -18, -27))
    )
    assert np.allclose(
        translate(test_coords_single, test_coords_single), np.array((0, 0, 0))
    )
    assert np.allclose(
        translate(test_coords_multi, np.array((1, 1, 1))), test_coords_multi - 1
    )

    assert np.allclose(
        translate(test_coords_single, np.array((10, 20, 30)), active=True),
        np.array((11, 22, 33)),
    )
    assert np.allclose(
        translate(test_coords_single, np.array((-1, -2, -3)), active=True),
        np.array((0, 0, 0)),
    )
    assert np.allclose(
        translate(test_coords_multi, np.array((1, 1, 1)), active=True),
        test_coords_multi + 1,
    )



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable = protected-access

import datetime

import pytest
from hypothesis import Verbosity, given, settings
from hypothesis.strategies import floats

import numpy as np

import pymedphys
import pymedphys._mocks.profiles
import pymedphys._wlutz.findfield
import pymedphys._wlutz.iview


def test_find_field_in_image():
    edge_lengths = [20, 20]

    expected_centre = [1.46, -1.9]
    expected_rotation = 12.1

    image_path = pymedphys.data_path("wlutz_image.png")
    x, y, img = pymedphys._wlutz.iview.iview_image_transform_from_path(image_path)

    _, centre, rotation = pymedphys._wlutz.core.find_field(x, y, img, edge_lengths)

    centre = np.round(centre, 2).tolist()
    rotation = float(np.round(rotation, 1))

    assert (expected_centre, expected_rotation) == (centre, rotation)


# @pytest.mark.slow
@pytest.mark.skip(
    reason=(
        "Wlutz field finding algorithm is being shelved for now. "
        "Will use pylinac only for the time being. When more resources "
        "are available, will implement independent algorithm to run "
        "alongside pylinac."
    )
)
@settings(
    deadline=datetime.timedelta(milliseconds=4000),
    max_examples=10,
    verbosity=Verbosity.verbose,
)
@given(
    floats(-20, 20),
    floats(-20, 20),
    floats(10, 20),
    floats(10, 20),
    floats(0.5, 3),
    floats(-360, 360),
)
def test_field_finding(x_centre, y_centre, x_edge, y_edge, penumbra, actual_rotation):
    edge_lengths = [x_edge, y_edge]
    actual_centre = [x_centre, y_centre]

    try:
        pymedphys._wlutz.findfield.check_aspect_ratio(edge_lengths)
    except ValueError:
        return

    field = pymedphys._mocks.profiles.create_rectangular_field_function(
        actual_centre, edge_lengths, penumbra, actual_rotation
    )

    x = np.arange(-50, 50, 0.1)
    y = np.arange(-50, 50, 0.1)
    xx, yy = np.meshgrid(x, y)
    zz = field(xx, yy)

    initial_centre = pymedphys._wlutz.findfield.get_centre_of_mass(x, y, zz)
    (centre, rotation) = pymedphys._wlutz.findfield.field_centre_and_rotation_refining(
        field, edge_lengths, penumbra, initial_centre
    )

    try:
        pymedphys._wlutz.findfield.check_rotation_and_centre(
            edge_lengths, actual_centre, centre, actual_rotation, rotation
        )
    except ValueError:
        print("Failed during comparison to gold reference values")
        raise


def test_find_initial_field_centre():
    centre = [20, 5]

    field = pymedphys._mocks.profiles.create_square_field_function(
        centre=centre, side_length=10, penumbra_width=1, rotation=20
    )

    x = np.arange(-15, 30, 0.1)
    y = np.arange(-15, 15, 0.1)

    xx, yy = np.meshgrid(x, y)

    zz = field(xx, yy)

    initial_centre = pymedphys._wlutz.findfield.get_centre_of_mass(x, y, zz)

    assert np.allclose(initial_centre, centre)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable = protected-access

from hypothesis import given
from hypothesis.strategies import floats

import numpy as np

import pymedphys._mocks.profiles
import pymedphys._wlutz.createaxis
import pymedphys._wlutz.interppoints


@given(
    floats(-20, 20),
    floats(-20, 20),
    floats(10, 20),
    floats(10, 20),
    floats(0, 2),
    floats(0, 360),
)
def test_transformed_field_interp(
    x_centre, y_centre, x_edge, y_edge, penumbra, degrees
):
    edge_lengths = [x_edge, y_edge]
    centre = [x_centre, y_centre]

    field = pymedphys._mocks.profiles.create_rectangular_field_function(
        centre, edge_lengths, penumbra, degrees
    )
    x_profile = pymedphys._mocks.profiles.create_profile_function(
        0, edge_lengths[0], penumbra
    )
    y_profile = pymedphys._mocks.profiles.create_profile_function(
        0, edge_lengths[1], penumbra
    )

    interp_size = np.max(edge_lengths) * 1.2

    field_x_interp = np.linspace(-interp_size / 2, interp_size / 2, 30)
    field_y_interp = np.linspace(-interp_size / 2, interp_size / 2, 40)

    transform = pymedphys._wlutz.interppoints.translate_and_rotate_transform(
        centre, degrees
    )

    x_interp, y_interp = pymedphys._wlutz.createaxis.transform_axis(
        field_x_interp, field_y_interp, transform
    )

    assert np.allclose(
        x_profile(field_x_interp), field(*x_interp), rtol=0.0001, atol=0.0001
    )
    assert np.allclose(
        y_profile(field_y_interp), field(*y_interp), rtol=0.0001, atol=0.0001
    )



import os
import pathlib
import subprocess

LIBRARY_ROOT = pathlib.Path(__file__).parent.parent.resolve()
PYLINT_RC_FILE = LIBRARY_ROOT.joinpath(".pylintrc")


def run_tests(_, remaining):
    original_cwd = os.getcwd()

    if "--pylint" in remaining:
        remaining.append(f"--pylint-rcfile={str(PYLINT_RC_FILE)}")

        if LIBRARY_ROOT.parent.name == "lib":
            working_directory_to_use = LIBRARY_ROOT.parent.parent
        else:
            working_directory_to_use = LIBRARY_ROOT.parent
    else:
        working_directory_to_use = LIBRARY_ROOT

    os.chdir(working_directory_to_use)
    print(f"Running tests with cwd set to:\n    {os.getcwd()}\n")

    try:
        command = " ".join(
            ["pytest"] + remaining + ["--pyargs", "pymedphys", "--failed-first"]
        )

        print(f"Running the following command:\n    {command}\n")

        subprocess.check_call(command, shell=True)
    finally:
        os.chdir(original_cwd)
